'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var createBabelFile = require('babel-file');
var createBabylonOptions = require('babylon-options');
var _typeof = require('@babel/runtime/helpers/typeof');
var _objectWithoutProperties = require('@babel/runtime/helpers/objectWithoutProperties');
var _objectSpread = require('@babel/runtime/helpers/objectSpread');
var nodePath = require('path');
var resolve = require('resolve');
var t = require('@babel/types');
var babelFileLoader = require('babel-file-loader');
var babelFlowIdentifiers = require('babel-flow-identifiers');
var babelTypeScopes = require('babel-type-scopes');
var babelIdentifiers = require('babel-identifiers');
var babelReactComponents = require('babel-react-components');
var babelNormalizeComments = require('babel-normalize-comments');
var babelExplodeModule = require('@aparna036/babel-explode-module');
var babelHelperSimplifyModule = require('babel-helper-simplify-module');
var printAST = require('ast-pretty-print');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var createBabelFile__default = /*#__PURE__*/_interopDefault(createBabelFile);
var createBabylonOptions__default = /*#__PURE__*/_interopDefault(createBabylonOptions);
var _typeof__default = /*#__PURE__*/_interopDefault(_typeof);
var _objectWithoutProperties__default = /*#__PURE__*/_interopDefault(_objectWithoutProperties);
var _objectSpread__default = /*#__PURE__*/_interopDefault(_objectSpread);
var nodePath__default = /*#__PURE__*/_interopDefault(nodePath);
var printAST__default = /*#__PURE__*/_interopDefault(printAST);

function hasDestructuredDefaultExport(path) {
  var exportPath = path.get('body').find(function (bodyPath) {
    return bodyPath.isExportNamedDeclaration() && bodyPath.get('specifiers').filter(function (n) {
      return n.node.exported.name === 'default';
    }).length;
  });
  return Boolean(exportPath);
}
function followExports(path, context, convert) {
  var exportPath = path.get('body').find(function (bodyPath) {
    return bodyPath.isExportNamedDeclaration() && bodyPath.get('specifiers').filter(function (n) {
      return n.node.exported.name === 'default';
    });
  });

  if (!exportPath) {
    throw new Error({
      message: 'No export path found'
    });
  }

  try {
    var filePath = babelFileLoader.resolveImportFilePathSync(exportPath, context.resolveOptions);
    var file = babelFileLoader.loadFileSync(filePath, context.parserOpts);
    var converted = convert(file.path, context);
    return converted;
  } catch (e) {
    throw new Error(e);
  }
}
function findExports(path, exportsToFind) {
  var formattedExports = [];
  path.get('body').filter(function (bodyPath) {
    return (// we only check for named and default exports here, we don't want export all
      exportsToFind === 'default' ? bodyPath.isExportDefaultDeclaration() : bodyPath.isExportNamedDeclaration() && bodyPath.node.source === null && ( // exportKind is 'value' or 'type' in flow
      bodyPath.node.exportKind === 'value' || // exportKind is undefined in typescript
      bodyPath.node.exportKind === undefined) || bodyPath.isExportDefaultDeclaration()
    );
  }).forEach(function (exportPath) {
    var declaration = exportPath.get('declaration');

    if (exportPath.isExportDefaultDeclaration()) {
      if (declaration.isIdentifier()) {
        var binding = path.scope.bindings[declaration.node.name].path;

        if (binding.isVariableDeclarator()) {
          binding = binding.get('init');
        }

        formattedExports.push({
          name: declaration.node.name,
          path: binding
        });
      } else {
        var name = null;

        if ((declaration.isClassDeclaration() || declaration.isFunctionDeclaration()) && declaration.node.id !== null) {
          name = declaration.node.id.name;
        }

        formattedExports.push({
          name: name,
          path: declaration
        });
      }
    } else {
      var specifiers = exportPath.get('specifiers');

      if (specifiers.length === 0) {
        if (declaration.isFunctionDeclaration() || declaration.isClassDeclaration()) {
          var identifier = declaration.node.id;
          formattedExports.push({
            name: identifier === null ? null : identifier.name,
            path: declaration
          });
        }

        if (declaration.isVariableDeclaration()) {
          declaration.get('declarations').forEach(function (declarator) {
            formattedExports.push({
              name: declarator.node.id.name,
              path: declarator.get('init')
            });
          });
        }
      } else {
        specifiers.forEach(function (specifier) {
          var name = specifier.node.local.name;
          var binding = path.scope.bindings[name].path;

          if (binding.isVariableDeclarator()) {
            binding = binding.get('init');
          }

          formattedExports.push({
            name: name,
            path: binding
          });
        });
      }
    }
  });
  return formattedExports;
}
function matchExported(file, exportName) {
  var exploded = babelExplodeModule.explodeModule(file.path.node);
  var statements = babelHelperSimplifyModule.explodedToStatements(exploded);
  var program = Object.assign({}, file.path.node, {
    body: statements
  });
  file.path.replaceWith(program);
  var match = exploded.exports.find(function (item) {
    return item.external === exportName;
  });

  if (!match) {
    return null;
  }

  var local = match.local;

  if (!local) {
    return null;
  }

  if (local === 'default' && match.source) {
    local = exportName;
  }

  var statement = file.path.get('body').find(function (item) {
    // Ignore export all & default declarations, since they do not have specifiers/ids.
    if (!item.isDeclaration() || item.isExportAllDeclaration()) return false;
    var id = null;

    if (item.isVariableDeclaration()) {
      id = item.node.declarations[0].id;
    } else if (item.isImportDeclaration()) {
      id = item.node.specifiers[0].local;
    } else if (item.isExportNamedDeclaration()) {
      id = item.node.specifiers[0].exported;
    } else if (item.node.id) {
      id = item.node.id;
    } else {
      throw new Error("Unexpected node:\n\n".concat(printAST__default['default'](item)));
    }

    if (!id) {
      throw new Error("Couldn't find id on node:\n\n".concat(printAST__default['default'](item)));
    }

    return id.name === local;
  });
  return statement || null;
}

var converters = {}; // This is the entry point. Program will only be found once.

converters.Program = function (path, context) {
  // coerce whether or not we need to follow an export to a new File and Program
  // only do so on export { default } from 'x';
  // followExports(path, context, convert);
  if (hasDestructuredDefaultExport(path)) {
    return followExports(path, context, convert);
  } else {
    var components = convertComponentExports(findExports(path, 'default'), context); // components[0] could be undefined

    var component;

    if (components[0]) {
      component = components[0].component;
    } // just extract the props from the first class in the file


    if (!component) {
      path.traverse({
        ClassDeclaration: function ClassDeclaration(scopedPath) {
          if (!component && babelReactComponents.isReactComponentClass(scopedPath)) {
            component = convertComponentClass(scopedPath, context);
          }
        }
      });
    }

    return {
      kind: 'program',
      component: component
    };
  }
};

converters.TaggedTemplateExpression = function (path, context) {
  return {
    kind: 'templateExpression',
    tag: convert(path.get('tag'), context)
  };
};

converters.TemplateElement = function (path) {
  return {
    kind: 'templateElement',
    value: path.node.value
  };
};

converters.TemplateLiteral = function (path, context) {
  return {
    kind: 'templateLiteral',
    expressions: path.get('expressions').map(function (e) {
      return convert(e, context);
    }),
    quasis: path.get('quasis').map(function (q) {
      return convert(q, context);
    })
  };
};

converters.LogicalExpression = function (path, context) {
  return {
    kind: 'logicalExpression',
    operator: path.node.operator,
    left: convert(path.get('left'), context),
    right: convert(path.get('right'), context)
  };
};

converters.RestElement = function (path, context) {
  return {
    kind: 'rest',
    argument: convert(path.get('argument'), context)
  };
};

converters.AssignmentPattern = function (path, context) {
  return {
    kind: 'assignmentPattern',
    left: convert(path.get('left'), context),
    right: convert(path.get('right'), context)
  };
};

converters.ObjectPattern = function (path, context) {
  var members = [];
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = path.get('properties')[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var property = _step.value;
      members.push(convert(property, context));
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator["return"] != null) {
        _iterator["return"]();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return {
    kind: 'objectPattern',
    members: members
  };
};

converters.ClassDeclaration = function (path, context) {
  if (!babelReactComponents.isReactComponentClass(path)) {
    return {
      kind: 'class',
      name: convert(path.get('id'), context)
    };
  }

  return convertComponentClass(path, context);
};

converters.SpreadElement = function (path, context) {
  return {
    kind: 'spread',
    value: convert(path.get('argument'), context)
  };
}; // Renamed to SpreadElement in babel 7. Added here for backwards compatibility


converters.SpreadProperty = function (path, context) {
  return {
    kind: 'spread',
    value: convert(path.get('argument'), context)
  };
};

converters.UnaryExpression = function (path, context) {
  return {
    kind: 'unary',
    operator: path.node.operator,
    argument: convert(path.get('argument'), context)
  };
};

converters.JSXAttribute = function (path, context) {
  return {
    kind: 'JSXAttribute',
    name: convert(path.get('name'), context),
    value: convert(path.get('value'), context)
  };
};

converters.JSXExpressionContainer = function (path, context) {
  return {
    kind: 'JSXExpressionContainer',
    expression: convert(path.get('expression'), context)
  };
};

converters.JSXElement = function (path, context) {
  return {
    kind: 'JSXElement',
    value: convert(path.get('openingElement'), context)
  };
};

converters.JSXIdentifier = function (path) {
  return {
    kind: 'JSXIdentifier',
    value: path.node.name
  };
};

converters.JSXMemberExpression = function (path, context) {
  return {
    kind: 'JSXMemberExpression',
    object: convert(path.get('object'), context),
    property: convert(path.get('property'), context)
  };
};

converters.JSXOpeningElement = function (path, context) {
  return {
    kind: 'JSXOpeningElement',
    name: convert(path.get('name'), context),
    attributes: path.get('attributes').map(function (item) {
      return convert(item, context);
    })
  };
};

converters.ClassProperty = function (path, context) {
  return {
    kind: 'property',
    key: convert(path.get('key'), context),
    value: convert(path.get('value'), context)
  };
};

converters.CallExpression = function (path, context) {
  var _convertCall = convertCall(path, context),
      callee = _convertCall.callee,
      args = _convertCall.args;

  return {
    kind: 'call',
    callee: callee,
    args: args
  };
};

converters.NewExpression = function (path, context) {
  var _convertCall2 = convertCall(path, context),
      callee = _convertCall2.callee,
      args = _convertCall2.args;

  return {
    kind: 'new',
    callee: callee,
    args: args
  };
};

converters.InterfaceDeclaration = function (path, context) {
  return {
    kind: 'interfaceDeclaration',
    id: convert(path.get('id'), context)
  };
};

converters.OpaqueType = function (path, context) {
  // OpaqueTypes have several optional nodes that exist as a null when not present
  // We need to convert these when they exist, and ignore them when they don't;
  var supertypePath = path.get('supertype');
  var impltypePath = path.get('impltype');
  var typeParametersPath = path.get('typeParameters'); // TODO we are having a fight at the moment with id returning a binding, not a node,
  // and don't have time to solve this properly - I am pathing it to being working-ish
  // here, and will come back to this later. If you find this comment still here and
  // want to fix this problem, I encourage you to do is.

  var supertype = supertypePath.node && convert(supertypePath, context);
  var impltype = impltypePath.node && convert(impltypePath, context);
  var typeParameters = typeParametersPath.node && convert(typeParametersPath, context);
  var id = convert(path.get('id'), context);
  return {
    kind: 'opaqueType',
    id: id,
    supertype: supertype,
    impltype: impltype,
    typeParameters: typeParameters
  };
};

converters.TypeofTypeAnnotation = function (path, context) {
  var type = convert(path.get('argument'), _objectSpread__default['default']({}, context, {
    mode: 'value'
  }));

  var _resolveFromGeneric = resolveFromGeneric(type),
      name = _resolveFromGeneric.name,
      referenceIdName = _resolveFromGeneric.referenceIdName;

  return {
    kind: 'typeof',
    type: type,
    name: name || referenceIdName
  };
};

converters.ObjectProperty = function (path, context) {
  return {
    kind: 'property',
    key: convert(path.get('key'), context),
    value: convert(path.get('value'), context)
  };
};

converters.ExistentialTypeParam = function () {
  return {
    kind: 'exists'
  };
};

converters.StringLiteral = function (path) {
  return {
    kind: 'string',
    value: path.node.value
  };
};

converters.TypeCastExpression = function (path, context) {
  return {
    kind: 'typeCastExpression',
    expression: convert(path.get('expression'), context)
  };
};

converters.NumericLiteral = function (path) {
  return {
    kind: 'number',
    value: path.node.value
  };
};

converters.NullLiteral = function () {
  return {
    kind: 'null'
  };
};

converters.BooleanLiteral = function (path) {
  return {
    kind: 'boolean',
    value: path.node.value
  };
};

converters.ArrayExpression = function (path, context) {
  return {
    kind: 'array',
    elements: path.get('elements').map(function (e) {
      return convert(e, context);
    })
  };
};

converters.BinaryExpression = function (path, context) {
  return {
    kind: 'binary',
    operator: path.node.operator,
    left: convert(path.get('left'), context),
    right: convert(path.get('right'), context)
  };
};

converters.MemberExpression = function (path, context) {
  return {
    kind: 'memberExpression',
    object: convert(path.get('object'), context),
    property: convert(path.get('property'), context)
  };
};

converters.FunctionDeclaration = function (path, context) {
  return convertFunction(path, context);
};

converters.ArrowFunctionExpression = function (path, context) {
  return convertFunction(path, context);
};

converters.FunctionExpression = function (path, context) {
  return convertFunction(path, context);
};

converters.TypeAnnotation = function (path, context) {
  return convert(path.get('typeAnnotation'), context);
};

converters.ExistsTypeAnnotation = function () {
  return {
    kind: 'exists'
  };
};

converters.ObjectTypeAnnotation = function (path, context) {
  return convertObject(path, context);
};

converters.ObjectTypeProperty = function (path, context) {
  return {
    kind: 'property',
    key: convert(path.get('key'), context),
    value: convert(path.get('value'), context),
    optional: path.node.optional
  };
};

converters.UnionTypeAnnotation = function (path, context) {
  return {
    kind: 'union',
    types: path.get('types').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.TypeParameterInstantiation = function (path, context) {
  return {
    kind: 'typeParams',
    params: path.get('params').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.TypeParameterDeclaration = function (path, context) {
  return {
    kind: 'typeParamsDeclaration',
    params: path.get('params').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.TypeParameter = function (path) {
  return {
    kind: 'typeParam',
    name: path.node.name
  };
};

converters.GenericTypeAnnotation = function (path, context) {
  var result = {};
  result.kind = 'generic';
  result.value = convert(path.get('id'), context);

  if (path.node.typeParameters) {
    result.typeParams = convert(path.get('typeParameters'), context);
  }

  if (result.value.kind === 'id') {
    result = convertUtilityTypes(result);
  }

  return result;
};

converters.ObjectMethod = function (path, context) {
  var parameters = path.get('params').map(function (p) {
    return convertParameter(p, context);
  });
  var returnType = path.node.returnType ? convert(path.get('returnType'), context) : null;
  return {
    kind: 'function',
    id: null,
    async: path.node.async,
    generator: path.node.generator,
    parameters: parameters,
    returnType: returnType
  };
};

converters.ObjectExpression = function (path, context) {
  return convertObject(path, context);
};

converters.VariableDeclaration = function (path, context) {
  return {
    kind: 'variable',
    declarations: path.get('declarations').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.VariableDeclarator = function (path, context) {
  return {
    kind: 'initial',
    id: convert(path.get('id'), context),
    value: convert(path.get('init'), context)
  };
};

converters.Identifier = function (path, context) {
  var kind = babelIdentifiers.getIdentifierKind(path);
  var name = path.node.name;

  if (context.mode === 'value') {
    if (kind === 'reference') {
      var binding = path.scope.getBinding(name);

      if (binding) {
        var bindingPath = binding.path;
        var foundPath = null;

        if (bindingPath.isVariableDeclaration()) {
          foundPath = bindingPath.get('declarators').find(function (p) {
            return p.node.name === name;
          });
        } else if (bindingPath.isVariableDeclarator()) {
          foundPath = bindingPath.get('init');
        } else if (bindingPath.isImportDefaultSpecifier() || bindingPath.isImportNamespaceSpecifier() || bindingPath.isImportSpecifier()) {
          foundPath = bindingPath;
        } else if (bindingPath.isDeclaration()) {
          foundPath = bindingPath.get('id');
        }

        if (foundPath === null || foundPath === undefined) {
          throw new Error("Unable to resolve binding path for: ".concat(bindingPath.type));
        }

        var convertedValue = convert(foundPath, context);
        return _objectSpread__default['default']({}, convertedValue, {
          referenceIdName: path.node.name
        });
      } else {
        var type = path.node.typeAnnotation ? convert(path.get('typeAnnotation'), _objectSpread__default['default']({}, context, {
          mode: 'type'
        })) : null;
        return {
          kind: 'id',
          name: name,
          type: type
        };
      }
    }

    if (kind === 'static' || kind === 'binding') {
      var _type = path.node.typeAnnotation ? convert(path.get('typeAnnotation'), _objectSpread__default['default']({}, context, {
        mode: 'type'
      })) : null;

      return {
        kind: 'id',
        name: name,
        type: _type
      };
    }

    throw new Error("Unable to resolve path for: ".concat(kind));
  }

  if (context.mode === 'type') {
    if (kind === 'reference') {
      var _bindingPath;

      if (babelFlowIdentifiers.isFlowIdentifier(path)) {
        var flowBinding = babelTypeScopes.getTypeBinding(path, name);
        if (!flowBinding) throw new Error();
        _bindingPath = flowBinding.path.parentPath;
      } else if (isTsIdentifier(path)) {
        var _foundPath = path.scope.getBinding(name);

        if (_foundPath && (_foundPath.path.isImportDefaultSpecifier() || _foundPath.path.isImportNamespaceSpecifier() || _foundPath.path.isImportSpecifier())) {
          return convert(_foundPath.path, context);
        }

        var tsBinding = babelTypeScopes.getTypeBinding(path, name);

        if (!tsBinding) {
          return {
            kind: 'id',
            name: name
          };
        }

        _bindingPath = tsBinding.path.parentPath;
      } else {
        _bindingPath = path.scope.getBinding(name);
      }

      if (_bindingPath) {
        if (_bindingPath.kind === 'module') {
          _bindingPath = _bindingPath.path;
        } // If path is a descendant of bindingPath and share the same name, this is a recursive type.


        if (path.isDescendant(_bindingPath) && _bindingPath.get('id').node.name === name) {
          return {
            kind: 'id',
            name: name
          };
        } // This is a hack that stops horrible regression errors and problems


        if (_bindingPath.kind === 'unknown') {
          return {
            kind: 'id',
            name: name
          };
        }

        if (_bindingPath.kind !== 'module') {
          var _convertedValue = convert(_bindingPath, context);

          return _objectSpread__default['default']({}, _convertedValue, {
            referenceIdName: path.node.name
          });
        }
      } else {
        return {
          kind: 'id',
          name: name
        };
      }
    }

    if (kind === 'static' || kind === 'binding') {
      return {
        kind: 'id',
        name: name
      };
    }
  }

  throw new Error("Could not parse Identifier ".concat(name, " in mode ").concat(context.mode));
};

converters.TypeAlias = function (path, context) {
  return convert(path.get('right'), context);
};

converters.IntersectionTypeAnnotation = function (path, context) {
  return {
    kind: 'intersection',
    types: path.get('types').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.QualifiedTypeIdentifier = function (path, context) {
  return convert(path.get('id'), context);
};

converters.VoidTypeAnnotation = function () {
  return {
    kind: 'void'
  };
};

converters.BooleanTypeAnnotation = function () {
  return {
    kind: 'boolean'
  };
};

converters.BooleanLiteralTypeAnnotation = function (path) {
  return {
    kind: 'boolean',
    value: path.node.value
  };
};

converters.NullLiteralTypeAnnotation = function () {
  return {
    kind: 'null'
  };
};

converters.StringLiteralTypeAnnotation = function (path) {
  return {
    kind: 'string',
    value: path.node.value
  };
};

converters.NumberLiteralTypeAnnotation = function (path) {
  return {
    kind: 'number',
    value: path.node.value
  };
};

converters.MixedTypeAnnotation = function () {
  return {
    kind: 'mixed'
  };
};

converters.AnyTypeAnnotation = function () {
  return {
    kind: 'any'
  };
};

converters.NumberTypeAnnotation = function () {
  return {
    kind: 'number'
  };
};

converters.FunctionTypeParam = function (path, context) {
  return convert(path.get('typeAnnotation'), context);
};

converters.FunctionTypeAnnotation = function (path, context) {
  return {
    parameters: path.get('params').map(function (p) {
      return convertParameter(p, context);
    }),
    returnType: convert(path.get('returnType'), context),
    kind: 'function'
  };
};

converters.StringTypeAnnotation = function () {
  return {
    kind: 'string'
  };
};

converters.NullableTypeAnnotation = function (path, context) {
  return {
    kind: 'nullable',
    arguments: convert(path.get('typeAnnotation'), context)
  };
};

converters.TSIndexedAccessType = function (path, context) {
  var type = convert(path.get('objectType'), context);
  var indexKey = path.get('indexType').node.literal.value;

  if (type.kind === 'generic') {
    if (type.value.members) {
      var member = type.value.members.find(function (scopedMember) {
        return scopedMember.key.name === indexKey;
      });

      if (member) {
        return member.value;
      }
    }

    var name = type.value.name || type.value.referenceIdName;
    return {
      kind: 'generic',
      value: {
        kind: type.value.kind,
        name: "".concat(name.name || name, "['").concat(indexKey, "']")
      }
    };
  }

  throw new Error("Unsupported TSIndexedAccessType kind: ".concat(type.kind));
};

converters.TSStringKeyword = function () {
  return {
    kind: 'string'
  };
};

converters.TSNumberKeyword = function () {
  return {
    kind: 'number'
  };
};

converters.TSBooleanKeyword = function () {
  return {
    kind: 'boolean'
  };
};

converters.TSVoidKeyword = function () {
  return {
    kind: 'void'
  };
};

converters.TSUndefinedKeyword = function () {
  return {
    kind: 'void'
  };
};

converters.TSTypeLiteral = function (path, context) {
  return {
    kind: 'object',
    members: path.get('members').map(function (memberPath) {
      return convert(memberPath, context);
    })
  };
};

converters.TSPropertySignature = function (path, context) {
  return {
    kind: 'property',
    optional: !!path.node.optional,
    key: {
      kind: 'id',
      name: path.node.key.name
    },
    value: convert(path.get('typeAnnotation'), context)
  };
};

converters.TSTypeAliasDeclaration = function (path, context) {
  return convert(path.get('typeAnnotation'), context);
};

converters.TSLiteralType = function (path) {
  return {
    kind: 'string',
    value: path.node.literal.value
  };
};

converters.TSTypeReference = function (path, context) {
  var typeParameters = path.get('typeParameters');
  return _objectSpread__default['default']({
    kind: 'generic',
    value: convert(path.get('typeName'), context)
  }, typeParameters.node && {
    key: convert(path.get('key'), context),
    typeParams: convert(typeParameters, context)
  });
};

converters.TSUnionType = function (path, context) {
  return {
    kind: 'union',
    types: path.get('types').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.TSAnyKeyword = function () {
  return {
    kind: 'any'
  };
};

converters.TSTupleType = function (path, context) {
  return {
    kind: 'tuple',
    types: path.get('elementTypes').map(function (p) {
      return convert(p, context);
    })
  };
};

converters.TSFunctionType = function (path, context) {
  return {
    kind: 'generic',
    value: {
      kind: 'function',
      returnType: convert(path.get('typeAnnotation'), context),
      parameters: path.get('parameters').map(function (p) {
        return convertParameter(p, context);
      })
    }
  };
};

function convertMethodCall(path, context) {
  var parameters = path.get('parameters').map(function (p) {
    return convertParameter(p, context);
  });
  var returnType = convert(path.get('typeAnnotation'), context);
  return {
    kind: 'function',
    returnType: returnType,
    parameters: parameters
  };
}

converters.TSMethodSignature = function (path, context) {
  return {
    kind: 'property',
    optional: !!path.node.optional,
    key: convert(path.get('key'), context),
    value: convertMethodCall(path, context)
  };
};

converters.TSCallSignatureDeclaration = function (path, context) {
  return {
    kind: 'property',
    key: {
      kind: 'string'
    },
    optional: false,
    value: convertMethodCall(path, context)
  };
};

function extendedTypesMembers(path, context) {
  var members = path.get('extends');

  if (!members || !members.length) {
    return [];
  }

  return members.reduce(function (acc, current) {
    var _convert = convert(current, context),
        convertedMembers = _convert.members; // #convertedMembers are undefined if converter is not able to resolve
    // extended types which are coming from some external package.


    if (!convertedMembers) {
      return acc;
    }

    return acc.concat(convertedMembers);
  }, []);
}

converters.TSInterfaceDeclaration = function (path, context) {
  var extendedTypes = extendedTypesMembers(path, context);
  var interfaceType = convert(path.get('body'), context) || {
    members: []
  };
  return {
    kind: 'object',
    // Merge the current interface members with any extended members
    members: interfaceType.members.concat(extendedTypes)
  };
};

converters.TSExpressionWithTypeArguments = function (path, context) {
  return convert(path.get('expression'), context);
};

converters.TSInterfaceBody = function (path, context) {
  return {
    kind: 'object',
    members: path.get('body').map(function (prop) {
      return convert(prop, context);
    })
  };
};

converters.TSTypeAnnotation = function (path, context) {
  return convert(path.get('typeAnnotation'), context);
};

converters.TSQualifiedName = function (path, context) {
  var left = convert(path.get('left'), context);
  var right = convert(path.get('right'), context);
  return {
    kind: 'id',
    name: "".concat(left.name || left.referenceIdName, ".").concat(right.name)
  };
};

converters.TSEnumDeclaration = function (path, context) {
  var name = path.get('id').node.name;
  var types = path.get('members').map(function (p) {
    var member = convert(p, context);
    return {
      kind: member.kind,
      name: "".concat(name, ".").concat(member.name)
    };
  });
  return {
    kind: 'union',
    types: types
  };
};

converters.TSEnumMember = function (path, context) {
  return convert(path.get('id'), context);
};

converters.TSArray = function () {
  return {
    kind: 'any'
  };
};

converters.TSArrayType = function (path, context) {
  return {
    kind: 'arrayType',
    type: convert(path.get('elementType'), context)
  };
};

converters.TSTypeParameterInstantiation = function (path, context) {
  return {
    kind: 'typeParams',
    params: path.get('params').map(function (param) {
      return convert(param, context);
    })
  };
};

converters.ImportNamespaceSpecifier = function () {
  return {
    kind: 'any'
  };
};

converters.undefined = function () {
  return {
    kind: 'any'
  };
};

converters.ObjectTypeSpreadProperty = function (path, context) {
  return {
    kind: 'spread',
    value: convert(path.get('argument'), context)
  };
};

converters.ArrayTypeAnnotation = function (path, context) {
  return {
    kind: 'arrayType',
    type: convert(path.get('elementType'), context)
  };
};

converters.TSIntersectionType = function (path, context) {
  return {
    kind: 'intersection',
    types: path.get('types').map(function (type) {
      return convert(type, context);
    })
  };
};

converters.TSIndexSignature = function (path, context) {
  var id = path.get('parameters')[0];
  return {
    kind: 'property',
    key: {
      kind: 'id',
      name: "[".concat(convert(id, context).name, ": ").concat(convert(id.get('typeAnnotation'), context).kind, "]")
    },
    value: convert(path.get('typeAnnotation'), context)
  };
};

converters.TSParenthesizedType = function (path, context) {
  return convert(path.get('typeAnnotation'), context);
};

converters.TSObjectKeyword = function () {
  return {
    kind: 'object',
    members: []
  };
};

converters.TSNullKeyword = function () {
  return {
    kind: 'null'
  };
};

converters.TSUnknownKeyword = function () {
  return {
    kind: 'unknown'
  };
};

converters.TSThisType = function () {
  return {
    kind: 'custom',
    value: 'this'
  };
};

converters.TSAsExpression = function (path, context) {
  return convert(path.get('expression'), context);
};

function importConverterGeneral(path, context) {
  var importKind = path.node.importKind || path.parent.importKind || 'value';
  var moduleSpecifier = path.parent.source.value;
  var name;
  var kind = path.parent.importKind;

  if (path.type === 'ImportDefaultSpecifier' && kind === 'value') {
    name = 'default';
  } else if (path.node.imported) {
    name = path.node.imported.name;
  } else {
    name = path.node.local.name;
  }

  if (!path.hub.file.opts.filename) {
    return {
      kind: 'import',
      importKind: importKind,
      name: name,
      moduleSpecifier: moduleSpecifier
    };
  } else {
    if (kind === 'typeof') {
      throw new Error({
        path: path,
        error: 'import typeof is unsupported'
      });
    }

    var filePath;

    try {
      filePath = babelFileLoader.resolveImportFilePathSync(path.parentPath, context.resolveOptions);
    } catch (e) {
      return {
        kind: 'import',
        importKind: importKind,
        name: name,
        moduleSpecifier: moduleSpecifier
      };
    }

    if (!filePath || nodePath__default['default'].extname(filePath) === '.json') {
      return {
        kind: 'import',
        importKind: importKind,
        name: name,
        moduleSpecifier: moduleSpecifier
      };
    }

    var file = babelFileLoader.loadFileSync(filePath, context.parserOpts);
    var id = path.node.imported ? path.node.imported.name : path.node.local.name;
    var exported = matchExported(file, name);

    if (!exported) {
      exported = recursivelyResolveExportAll(file.path, context, name);

      if (!exported) {
        return {
          kind: 'import',
          importKind: importKind,
          name: name,
          moduleSpecifier: moduleSpecifier
        };
      }
    }

    return convert(exported, _objectSpread__default['default']({}, context, {
      replacementId: t.identifier(id)
    }));
  }
}

converters.ImportDefaultSpecifier = function (path, context) {
  return importConverterGeneral(path, context);
};

converters.ImportDeclaration = function (path, context) {
  var filePath = babelFileLoader.resolveImportFilePathSync(path, context.resolveOptions);
  var file = babelFileLoader.loadFileSync(filePath, context.parserOpts);
  var exported = matchExported(file, context.replacementId.name);
  var importKind = path.node.importKind || 'value';
  var moduleSpecifier = path.get('source').node.value;
  var name = 'default';

  if (!exported || !context.replacementId) {
    return {
      kind: 'import',
      importKind: importKind,
      name: name,
      moduleSpecifier: moduleSpecifier
    };
  }

  return convert(exported, context);
};

converters.ExportSpecifier = function (path, context) {
  return {
    kind: 'exportSpecifier',
    local: convert(path.get('local'), context),
    exported: convert(path.get('exported'), context)
  };
};

converters.ExportNamedDeclaration = function (path, context) {
  var specifiers = path.get('specifiers'); // This needs to be in all of them --- let source = path.get('source');

  if (path.get('source').node) {
    var source = path.get('source');

    if (specifiers.length !== 1) {
      return {
        kind: 'export',
        exports: specifiers.map(function (s) {
          return convert(s, context);
        }),
        source: convert(source, context)
      };
    }

    try {
      // The parentPath is a reference to where we currently are. We want to
      // get the source value, but resolving this first makes this easier.
      var filePath = babelFileLoader.resolveImportFilePathSync(source.parentPath, context.resolveOptions);
      var actualPath = resolve.sync(nodePath__default['default'].join(nodePath__default['default'].dirname(filePath), source.node.value), context.resolveOptions);
      var name = convert(specifiers[0], context).local.name;
      var file = babelFileLoader.loadFileSync(actualPath, context.parserOpts); // We need to calculate name from the specifiers, I think knowing that there
      // will always be one specifier

      var resolvedValue = matchExported(file, name);

      if (resolvedValue) {
        return convert(resolvedValue, context);
      }

      return {
        kind: 'export',
        exports: specifiers.map(function (s) {
          return convert(s, context);
        }),
        source: convert(source, context)
      };
    } catch (e) {
      return {
        kind: 'export',
        exports: specifiers.map(function (s) {
          return convert(s, context);
        }),
        source: convert(source, context)
      };
    }
  } else {
    return {
      kind: 'export',
      exports: specifiers.map(function (s) {
        return convert(s, context);
      })
    };
  }
};

converters.ImportSpecifier = function (path, context) {
  return importConverterGeneral(path, context);
};

converters.TSConditionalType = function () {
  return {
    kind: 'any'
  };
};

converters.TSTypeQuery = function (path, context) {
  return {
    kind: 'typeQuery',
    exprName: convert(path.get('exprName'), _objectSpread__default['default']({}, context, {
      mode: 'value'
    }))
  };
};

function attachCommentProperty(source, dest, name) {
  if (!source || !source[name]) return;
  if (!dest[name]) dest[name] = [];
  dest[name] = dest[name].concat(parseComment(source[name]));
}

function attachComments(source, dest) {
  attachCommentProperty(source, dest, 'leadingComments');
  attachCommentProperty(source, dest, 'trailingComments');
  attachCommentProperty(source, dest, 'innerComments');
}

function parseComment(commentProperty) {
  return commentProperty.map(function (comment) {
    return {
      type: comment.type === 'CommentLine' ? 'commentLine' : 'commentBlock',
      value: babelNormalizeComments.normalizeComment(comment),
      raw: comment.value
    };
  });
}

var isSpecialComponentType = function isSpecialComponentType(path, type) {
  if (path && path.isCallExpression()) {
    var callee = path.get('callee');

    if (callee.isIdentifier() && callee.node.name === type) {
      return true;
    }

    if (callee.isMemberExpression() && callee.matchesPattern("React.".concat(type))) {
      return true;
    }
  }

  return false;
};

function convertObject(path, context) {
  var members = [];
  path.get('properties').forEach(function (p) {
    var mem = convert(p, context);

    if (mem.kind === 'spread') {
      var memVal = resolveFromGeneric(mem.value);

      if (memVal.kind === 'initial' && memVal.value.kind === 'object') {
        members = members.concat(memVal.value.members);
      } else if (memVal.kind === 'object') {
        members = members.concat(memVal.members);
      } else if (memVal.kind === 'variable') {
        var declarations = memVal.declarations;
        declarations = declarations[declarations.length - 1].value;

        if (declarations.kind !== 'object') {
          throw new Error('Trying to spread a non-object item onto an object');
        } else {
          members = members.concat(declarations.members);
        }
      } else if (memVal.kind === 'import') {
        // We are explicitly calling out we are handling the import kind
        members = members.concat(mem);
      } else {
        // This is a fallback
        members = members.concat(mem);
      }
    } else if (mem.kind === 'property') {
      members.push(mem);
    }
  });
  return {
    kind: 'object',
    members: members
  };
}

function resolveExportAllDeclaration(path, context) {
  var source = path.get('source'); // The parentPath is a reference to where we currently are. We want to
  // get the source value, but resolving this first makes this easier.

  var filePath = babelFileLoader.resolveImportFilePathSync(source.parentPath, context.resolveOptions);
  return babelFileLoader.loadFileSync(filePath, context.parserOpts);
} // Converts utility types to a simpler representation


function convertUtilityTypes(type) {
  var result = _objectSpread__default['default']({}, type);

  if (type.value.name === '$Exact') {
    // $Exact<T> can simply be converted to T
    if (type.typeParams && type.typeParams.params && type.typeParams.params[0]) {
      result = type.typeParams.params[0];
    } else {
      /* eslint-disable-next-line no-console */
      console.warn('Missing type parameter for $Exact type');
    }
  }

  return result;
}

function convertComponentClass(path, context) {
  var params = path.get('superTypeParameters').get('params');
  var props = params[0];
  var defaultProps = getDefaultProps(path, context);
  var classProperties = convert(props, _objectSpread__default['default']({}, context, {
    mode: 'type'
  }));
  classProperties.name = convert(path.get('id'), _objectSpread__default['default']({}, context, {
    mode: 'value'
  }));
  /**
   * FIXME: It's possible to get nulls in the members array when TS is unable
   * to resolve type definitions of non-relative module imports
   * See: https://github.com/atlassian/extract-react-types/issues/89
   **/

  if (classProperties.value && classProperties.value.members) {
    classProperties.value.members = classProperties.value.members.filter(function (m) {
      return !!m;
    });
  }

  return addDefaultProps(classProperties, defaultProps);
}

function convertComponentFunction(path, context, propTypes) {
  // we have a function, assume the props are the first parameter
  var functionProperties = convert(propTypes, _objectSpread__default['default']({}, context, {
    mode: 'type'
  }));
  var name = '';

  if (path.type === 'FunctionDeclaration' && path.node.id && path.node.id.name) {
    name = path.node.id.name;
  } else {
    var variableDeclarator = path.findParent(function (scopedPath) {
      return scopedPath.isVariableDeclarator();
    });

    if (variableDeclarator) {
      name = variableDeclarator.node.id.name;
    }
  }

  var defaultProps = [];

  if (name) {
    path.hub.file.path.traverse({
      // look for MyComponent.defaultProps = ...
      AssignmentExpression: function AssignmentExpression(assignmentPath) {
        var left = assignmentPath.get('left.object');

        if (left.isIdentifier() && left.node.name === name) {
          var initialConversion = convert(assignmentPath.get('right'), _objectSpread__default['default']({}, context, {
            mode: 'value'
          }));
          defaultProps = initialConversion.members;
        }
      }
    });
    functionProperties.name = {
      kind: 'id',
      name: name,
      type: null
    };
  }

  return addDefaultProps(functionProperties, defaultProps);
}

function addDefaultProps(props, defaultProps) {
  if (!defaultProps) return props;
  defaultProps.forEach(function (property) {
    var ungeneric = resolveFromGeneric(props);
    var prop = getProp(ungeneric, property);

    if (!prop) {
      /* eslint-disable-next-line no-console */
      console.warn("Could not find property to go with default of ".concat(property.key.value ? property.key.value : property.key.name, " in ").concat(props.name, " prop types"));
      return;
    }

    prop["default"] = property.value;
  });
  return props;
}

function convertCall(path, context) {
  var callee = convert(path.get('callee'), context);
  var args = path.get('arguments').map(function (a) {
    return convert(a, context);
  });
  return {
    callee: callee,
    args: args
  };
}

function recursivelyResolveExportAll(path, context, name) {
  var source = path.get('body').filter(function (item) {
    return item.isExportAllDeclaration();
  }).map(function (item) {
    return resolveExportAllDeclaration(item, context);
  }).filter(Boolean);
  var matchedDeclartion = source.reduce(function (acc, current) {
    if (acc) {
      return acc;
    }

    return matchExported(current, name);
  }, null);

  if (matchedDeclartion) {
    return matchedDeclartion;
  }

  return source.reduce(function (acc, current) {
    if (acc) {
      return acc;
    }

    return recursivelyResolveExportAll(current.path, context, name);
  }, null);
}

var getPropFromObject = function getPropFromObject(props, property) {
  var prop;

  if (!props.members) {
    throw new Error("Attempted to get property from non-object kind: ".concat(props.kind, ". Full object: ").concat(JSON.stringify(props)));
  }

  props.members.forEach(function (p) {
    if (p.kind === 'spread') {
      var spreadArg = resolveFromGeneric(p.value);
      if (!spreadArg || spreadArg.kind !== 'object') return;
      var p2 = getPropFromObject(spreadArg, property);
      if (p2) prop = p2; // The kind of the object member must be the same as the kind of the property
    } else if (property.key.kind === 'id' && p.key.name === property.key.name) {
      prop = p;
    } else if (property.key.kind === 'string' && p.key.value === property.key.value) {
      prop = p;
    }
  });
  return prop;
};

var resolveFromGeneric = function resolveFromGeneric(type) {
  if (type.kind !== 'generic') return type;
  return resolveFromGeneric(type.value);
};

var getProp = function getProp(props, property) {
  var prop;

  if (props.kind === 'intersection') {
    props.types.forEach(function (pr) {
      prop = getProp(resolveFromGeneric(pr), property) || prop;
    });
  } else if (props.kind === 'object') {
    prop = getPropFromObject(props, property);
  }

  return prop;
};

var isVariableOfMembers = function isVariableOfMembers(defaultProps) {
  var defaultPropsIsVar = defaultProps.value && defaultProps.value.kind === 'variable';
  if (!defaultPropsIsVar) return false;
  var declarations = defaultProps.value.declarations;
  var lastDeclarationIsObject = declarations[declarations.length - 1].value.kind === 'object';
  return lastDeclarationIsObject;
};

var getDefaultProps = function getDefaultProps(path, context) {
  var defaultProps = null;
  var foundDefaults = path.get('body').get('body').find(function (p) {
    return p.isClassProperty() && p.get('key').isIdentifier({
      name: 'defaultProps'
    });
  });

  if (foundDefaults) {
    defaultProps = convert(foundDefaults, _objectSpread__default['default']({}, context, {
      mode: 'value'
    }));
  }

  if (!defaultProps) return [];

  if (defaultProps && defaultProps.value && defaultProps.value.kind === 'object') {
    return defaultProps.value.members;
  }

  if (isVariableOfMembers(defaultProps)) {
    return defaultProps.value.declarations[defaultProps.value.declarations.length - 1].value.members;
  }

  throw new Error("Could not resolve default Props, ".concat(defaultProps));
};

function isTsIdentifier(path) {
  return ['TSExpressionWithTypeArguments', 'TSTypeReference'].indexOf(path.parentPath.type) !== -1 && babelIdentifiers.getIdentifierKind(path) === 'reference';
}

function convertParameter(param, context) {
  var _convert2 = convert(param, context),
      type = _convert2.type,
      rest = _objectWithoutProperties__default['default'](_convert2, ["type"]);

  return {
    kind: 'param',
    value: rest,
    type: type || null
  };
}

function convertFunction(path, context) {
  var parameters = path.get('params').map(function (p) {
    return convertParameter(p, context);
  });
  var returnType = path.node.returnType ? convert(path.get('returnType'), context) : null;
  var id = path.node.id ? convert(path.get('id'), context) : null;
  return {
    kind: 'function',
    id: id,
    async: path.node.async,
    generator: path.node.generator,
    parameters: parameters,
    returnType: returnType
  };
} // This function is from mdn:
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Errors/Cyclic_object_value#Examples


var getCircularReplacer = function getCircularReplacer() {
  var seen = new WeakSet();
  return function (key, value) {
    if (_typeof__default['default'](value) === 'object' && value !== null) {
      if (seen.has(value)) {
        return;
      }

      seen.add(value);
    }

    return value;
  };
};

function convert(path, context) {
  if (typeof path.get !== 'function') {
    // We were getting incredible unhelpful errors here at times, so we have a circular replacement
    // throw path.identifier;
    var stringedPath = JSON.stringify(path, getCircularReplacer(), 2);
    throw new Error("Did not pass a NodePath to convert() ".concat(stringedPath));
  }

  var converter = converters[path.type];

  if (!converter) {
    var propertySignature = path.find(function (p) {
      return p.isTSPropertySignature() || p.isObjectTypeProperty();
    }); // Fallback to a raw string if property uses a type without a matching converter

    if (propertySignature && propertySignature.node) {
      return {
        kind: 'raw',
        name: path.getSource()
      };
    }

    throw new Error("Missing converter for: ".concat(path.type, ", see: https://github.com/atlassian/extract-react-types/issues/150"));
  }

  var result = converter(path, context);
  attachComments(path.node, result);
  return result;
}
function convertComponentExports(componentExports, context) {
  // eslint-disable-next-line array-callback-return
  return componentExports.map(function (_ref) {
    var path = _ref.path,
        name = _ref.name;

    if (path.isFunctionExpression() || path.isArrowFunctionExpression() || path.isFunctionDeclaration()) {
      var propType; // check for a component typed with the `FC<Props>` annotation

      if (path.parentPath.node.id && path.parentPath.node.id.typeAnnotation && path.parentPath.node.id.typeAnnotation.typeAnnotation.typeName.name === 'FC' && path.parentPath.node.id.typeAnnotation.typeAnnotation.typeParameters) {
        propType = path.parentPath.get('id.typeAnnotation.typeAnnotation.typeParameters.params.0');
      } else {
        // we have a normal function, assume the props are the first parameter
        propType = path.get('params.0.typeAnnotation');
      }

      var component = convertComponentFunction(path, context, propType);
      return {
        name: name,
        path: path,
        component: component
      };
    }

    if (path.isClass()) {
      var _component = convertComponentClass(path, context);

      return {
        name: name,
        path: path,
        component: _component
      };
    }

    var isMemo = isSpecialComponentType(path, 'memo');
    var isForwardRef = isSpecialComponentType(path, 'forwardRef');

    if (isMemo || isForwardRef) {
      // Props typed via generics
      var genericParams = path.get('typeParameters'); // Props are the second type arg

      if (isForwardRef && genericParams.node) {
        var _component2 = convertComponentFunction(genericParams, context, genericParams.get('params.1'));

        return {
          name: name,
          path: path,
          component: _component2
        };
      }

      if (isMemo && genericParams.node) {
        var _component3 = convertComponentFunction(genericParams, context, genericParams.get('params.0'));

        return {
          name: name,
          path: path,
          component: _component3
        };
      } // Props typed via function arguments


      var firstArg = path.get('arguments')[0];

      if (firstArg) {
        if (firstArg.isFunctionExpression() || firstArg.isArrowFunctionExpression()) {
          var _component4 = convertComponentFunction(firstArg, context, firstArg.get('params.0.typeAnnotation'));

          return {
            name: name,
            path: path,
            component: _component4
          };
        }

        var firstArgGenericParams = firstArg.get('typeParameters');

        if (isMemo && isSpecialComponentType(firstArg, 'forwardRef') && firstArgGenericParams.node) {
          var _component5 = convertComponentFunction(firstArgGenericParams, context, firstArgGenericParams.get('params.1'));

          return {
            name: name,
            path: path,
            component: _component5
          };
        }

        if (isMemo && isSpecialComponentType(firstArg, 'forwardRef')) {
          var innerFirstArg = firstArg.get('arguments')[0];

          if (innerFirstArg.isFunctionExpression() || innerFirstArg.isArrowFunctionExpression()) {
            var _component6 = convertComponentFunction(innerFirstArg, innerFirstArg, innerFirstArg.get('params.0.typeAnnotation'));

            return {
              name: name,
              path: path,
              component: _component6
            };
          }
        }
      }
    }
  });
}

function getContext(typeSystem, filename) {
  var resolveOptions = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  var plugins = ['jsx', ['decorators', {
    decoratorsBeforeExport: true
  }]];

  if (!resolveOptions.extensions) {
    // The resolve package that babel-file-loader uses only resolves .js files by default instead of the
    // default extension list of node (.js, .json and .node) so add .json back here.
    resolveOptions.extensions = ['.js', '.json'];
  }

  if (typeSystem !== 'flow' && typeSystem !== 'typescript') {
    throw new Error('typeSystem must be either "flow" or "typescript"');
  }

  if (typeSystem === 'flow') {
    plugins.push(['flow', {
      all: true
    }]);
  }

  if (typeSystem === 'typescript') {
    plugins.push('typescript');
    resolveOptions.extensions.push('.tsx');
    resolveOptions.extensions.push('.ts');
  }
  /* $FlowFixMe - need to update types in babylon-options */


  var parserOpts = createBabylonOptions__default['default']({
    stage: 2,
    plugins: plugins
  });
  return {
    resolveOptions: resolveOptions,
    parserOpts: parserOpts
  };
}

function extractReactTypes(code, typeSystem, filename, inputResolveOptions) {
  var _getContext = getContext(typeSystem, filename, inputResolveOptions),
      resolveOptions = _getContext.resolveOptions,
      parserOpts = _getContext.parserOpts;

  var file = createBabelFile__default['default'](code, {
    parserOpts: parserOpts,
    filename: filename
  });
  return convert(file.path, {
    resolveOptions: resolveOptions,
    parserOpts: parserOpts
  });
}
function findExportedComponents(programPath, typeSystem, filename, resolveOptions) {
  return convertComponentExports(findExports(programPath, 'all'), getContext(typeSystem, filename, resolveOptions));
}

exports.extractReactTypes = extractReactTypes;
exports.findExportedComponents = findExportedComponents;
